// Copyright(c) 2014, The LIMIX developers(Christoph Lippert, Paolo Francesco Casale, Oliver Stegle)
//
//Licensed under the Apache License, Version 2.0 (the "License");
//you may not use this file except in compliance with the License.
//You may obtain a copy of the License at
//
//    http://www.apache.org/licenses/LICENSE-2.0
//
//Unless required by applicable law or agreed to in writing, software
//distributed under the License is distributed on an "AS IS" BASIS,
//WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
//See the License for the specific language governing permissions and
//limitations under the License.

#ifndef CDATA_H_
#define CDATA_H_

#include "ADataTerm.h"

namespace limix {


class CData: public ADataTerm
{
public:
	CData();
	CData(MatrixXd& Y);
	~CData();
	virtual inline std::string getName() const {return "CData";};
	muint_t getRowsParams() {return 0;};
	muint_t getColsParams() {return 0;};
};



} /* namespace limix */
#endif /* CDATA_H_ */
